package plaGen;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.Stack;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;



public class Caseinfo {
	private int numInputBits;
	private BufferedReader vfile;
	private int numOutputBits;
	private int numCases;
	private int[][] inputMatrix;
	private int[][] outputMatrix;
	private String inBusName;
	private String outBusName;

	public Caseinfo() {
		numInputBits=-1;
		numOutputBits=-1;
		
		fileIn();
		Parser();
	}

	private void Parser() {
		String str = null;
		
		//Get the top line of the case statement
		try {
			str = vfile.readLine();
		} catch (IOException e) {
			e.printStackTrace();
		}

    	while(str != null){
    		System.out.println("Parsing " + str);

    		if( CaseFound(str) ) {
    			//System.out.println("Case Found");
    			ParseCaseHeader(str);
    			try {
					ParseCaseBody();
				} catch (IOException e) {
					e.printStackTrace();
				}
    		}
    		
    		try {
				str = vfile.readLine();
			} catch (IOException e) {
				e.printStackTrace();
			}      		
    	}
	}

	private void ParseCaseHeader(String str) {
		int startparens = str.indexOf("(");
		int endparens = str.indexOf(")");
		inBusName = str.substring(startparens+1,endparens);
	}

	private void ParseCaseBody() throws IOException{
    	String input;
    	String output;
    	boolean isWhite;
    	boolean isEndCase;
    	//boolean isComInput;
    	boolean isComment = false;
    	boolean isDefault = true;
    	boolean hasDefault = false;
    	Stack outputStack = new Stack();
    	Stack inputStack = new Stack();
    	String str=null;
    	Pattern inBitsPat  = Pattern.compile("^\\s*(\\d+)");
    	Pattern outBitsPat = Pattern.compile("<=\\s*(\\d+)");
		Pattern whitePat   = Pattern.compile("^\\s*$");
		Pattern endCasePat = Pattern.compile("^\\s*endcase");
		Pattern commentPat = Pattern.compile("^\\s*//");
		Pattern defaultPat = Pattern.compile("^\\s*default");
		//Pattern comInputPat = Pattern.compile("^\\s*d+'b");
		Pattern outBus = Pattern.compile(":\\s*(\\w+)");
    	
    	
    	try {
			str = vfile.readLine();
		} catch (IOException e) {
			e.printStackTrace();
		}   

    	while(str!= null){
    		
			Matcher matcher;
			//System.out.println("PCB parsing"+str);
			
	    	isWhite = whitePat.matcher(str).find();
	    	//System.out.println("isWhite"+isWhite);
	    	
	    	isEndCase = endCasePat.matcher(str).find();
	    	//System.out.println("isEndCase"+isEndCase);
	    	
	    	isComment = commentPat.matcher(str).find();
	    	//System.out.println("isComment ="+isComment);
	    	
	    	//isComInput = comInputPat.matcher(str).find();
	    	
	    	isDefault = defaultPat.matcher(str).find();
	    	//System.out.println("isDefault ="+isDefault);
	    	
	    	// EndCase scenario
	        if(isEndCase) {
	        	//System.out.println("isEndCase = True");
	        	if(!hasDefault){
	        		throw new IOException("Must declare proper default statement");
	        	}

	        	if(inputStack.size() != outputStack.size()){
	        		throw new IOException("Number of input and output cases should match");
	        	}
	        	
	        	int inSz  = inputStack.size();
	        	int outSz = outputStack.size();
	        	input = (String)inputStack.pop();
	        	output = (String)outputStack.pop();	
	        	int inLn = input.length();
	        	int outLn = output.length();

	        	numCases=inSz;
	        	
	        	//System.out.println("inSz = " +inSz+ " inLn = " +inLn+ " outSz = " +outSz+ " outLn = "+outLn);

	        	inputMatrix = new int [inSz][inLn];
	        	outputMatrix = new int [outSz][outLn];

	        	for(int j = 0; j < inLn; j++){
	        		if(input.charAt(j)=='0'){
	        			inputMatrix[inSz-1][j] = 0;
	        		}
	        		else if(input.charAt(j)=='1'){
	        			inputMatrix[inSz-1][j] = 1;
	        		}
	        		else{
	        			inputMatrix[inSz-1][j] = 2;
	        		}
	        	}

				for(int i = 1; i < inSz; i++) {
					input = (String)inputStack.pop();
					

					for(int j = 0; j < inLn; j++){
						if(input.charAt(j)=='0'){
							inputMatrix[inSz-(1+i)][j] = 0;
		        		}
		        		else if(input.charAt(j)=='1'){
		        			inputMatrix[inSz-(1+i)][j] = 1;
		        		}
		        		else{
		        			inputMatrix[inSz-(1+i)][j] = 2;
		        		}
					}
				}

				for(int j = 0; j < outLn; j++){
					outputMatrix[outSz-1][j] = Character.digit(output.charAt(j),10);
				}

				for(int i = 1; i < outSz; i++) {
					output = (String)outputStack.pop();

					for(int j = 0; j < outLn; j++){
						outputMatrix[outSz-(1+i)][j] = Character.digit(output.charAt(j),10);
					}
				}
				
				return;
	        }

	        else if( (!isWhite) && (!isComment) && (!isDefault) ){
	        	//System.out.println("Regular");
	    	   	// Find number of bits on input
	    	   	matcher = inBitsPat.matcher(str);

		    	if(matcher.find()){
		    		if(numInputBits == -1) {
		    			numInputBits = Integer.valueOf(matcher.group(1));
		    		}
		    		
		    		else {
		    			if(numInputBits != Integer.valueOf(matcher.group(1))) {
		    				throw new IOException("Number of input bits not consistent in case statement");
		    			}
		    		}
		    	}

		    	else {
		    		throw new IOException("Invalid Case while looking for num input bits");
		    	}

		    	// Find number of bits on ouput
		    	matcher = outBitsPat.matcher(str);

		    	if(matcher.find()) {
		    		if(numOutputBits == -1) {
		    			numOutputBits = Integer.valueOf(matcher.group(1));
		    			Matcher newmatcher = outBus.matcher(str);
		    			if (newmatcher.find()){
		    				outBusName=newmatcher.group(1);
		    			}
		    			else throw new IOException("Invalid case statement format.");
		    		}

		    		else {
		    			if(numOutputBits != Integer.valueOf(matcher.group(1))) {
		    				throw new IOException("Number of output bits not consistent in case statement");
		    			}
		    		}
		    	}

		    	else {
		    		throw new IOException("Invalid Case while looking for num output bits");
		    	}

		    	// Create patterns for searching for input and output
		    	Pattern inPat = Pattern.compile("b([01?]{"+String.valueOf(numInputBits)+"})\\s*:");
				Pattern outPat = Pattern.compile("b(\\d{"+String.valueOf(numOutputBits)+"})\\s*;");

		    	// Find input
		    	matcher = inPat.matcher(str);

		    	if(matcher.find()){
		    		input = matcher.group(1);
		    	}

		    	else{ 
		    		throw new IOException("Could not find matching input");
		    	}

		    	// Find output
		    	matcher = outPat.matcher(str);

		    	if(matcher.find()){
		    		output = matcher.group(1);
		    	}

		    	else{
		    		throw new IOException("Could not find matching output");
		    	}
		    	
		    	inputStack.push(input);
		    	outputStack.push(output);
	    	}

	    	else if(isDefault) {
	    		//System.out.println("isDefault = True");
	    		Pattern defPat = Pattern.compile("'bx+");

	    		if(defPat.matcher(str).find()){
	    			hasDefault = true;
	    		}
	    	}
	        //System.out.println("reading next line");
    		str = vfile.readLine();
    	}
    }



	public static boolean CaseFound(String str){
    	Pattern pattern = Pattern.compile("case");
    	Matcher matcher = pattern.matcher(str);
    	boolean found = false;
    	
    	while (matcher.find()){
    		found = true;
    	}
    	
    	return found;
    }



	private void fileIn() {
		JOptionPane.showMessageDialog(null, "Now choose the Case-File (.v)",
				"File-Prompt",JOptionPane.PLAIN_MESSAGE);
		try{
			JFileChooser filechooser = new JFileChooser();
			int result = filechooser.showOpenDialog(null);
			if(result == JFileChooser.CANCEL_OPTION) System.exit(1);
				
			//  Imports case-file
			vfile = new BufferedReader(new FileReader(filechooser.getSelectedFile()));
		}
		catch(IOException e){
			System.out.println(e);
		}
	}
	public int getNumInputBits(){
		return numInputBits;
	}
	
	public int getNumOutputBits(){
		return numOutputBits;
	}
	
	public int getNumCases(){
		return numCases;
	}
	
	public int[][] getInputMatrix(){
		return inputMatrix;
	}
	
	public int[][] getOutputMatrix(){
		return outputMatrix;
	}

	public String getInBusName() {
		return inBusName;
	}

	public String getOutBusName() {
		return outBusName;
	}
}
