package plaGen;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.StringTokenizer;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;

/*	
 * This class takes is used in the plaGEN tool to obtain and make
 * available information regarding cell parameters for a VLSI build.
 * As of the first version, a plain-text file is imported and parsed to
 * obtain the necessisary information.  It is anticipated that better
 * functionality might be achieved by using a GUI of some sort to store
 * information more "cleanly",instead of storing the information in the 
 * text file.
 * All parameters are available to be called using standard "getters".
 */
public class Paramgetter {

	private String cellName;// name to which the output file will be stored

	private String eleVersion;// version of Electric that is used.

	private double m2Width;// width of the Metal-2 arcs

	private double m1Width;// width of the Metal-1 arcs

	private double polyWidth;// width of polysilicon arcs

	private double activeWidth;

	private double weakActiveWidth;

	private double weakPmosSize;

	private double weakPolyWidth;// width of "weak" polysilicon arcs

	private double powerAndGndWide;// width of power and ground arcs

	private double m2ToM2;// minimum distance seperating metal-2 arcs

	private double m1ToM1;// minimum distance seperating metal-1 arcs

	private double polyToPoly;// minimum distance seperating polysilicon arcs

	private double polyToDiff;// minimum distance between polysilicon and

	// diffusion

	private double wellSurround;// distance of well buffer surrounding xistors.

	private double transistorToActiveConnection;// minimum distance between

	// xistors

	// and M1 active connections.
	private double weakTransistorToActiveConnection;// minimum distance beween

	// "weak" xistors and m1 active connections.

	private double nMosInvWidth;// size of the input and output inverters for

	// the PLA

	public Paramgetter() {
		getParameters();
	}

	/*
	 * This method reads in the plain-text file and pulls out pertenant data. It
	 * is not very robust, in that any deviation from the file format, (such as
	 * blank lines, etc...) will result in an error. Only works if the file
	 * format and layout is exactly as specified, though the actual order of the
	 * parameter lines in the file should not matter.
	 */
	private void getParameters() {
		JOptionPane.showMessageDialog(null,
				"Please choose the Parameter-File (.txt)", "File-Prompt",
				JOptionPane.PLAIN_MESSAGE);
		try {
			JFileChooser filechooser = new JFileChooser();
			double result = filechooser.showOpenDialog(null);
			if (result == JFileChooser.CANCEL_OPTION)
				System.exit(1);

			// Imports parameter file
			BufferedReader fileIn = new BufferedReader(new FileReader(
					filechooser.getSelectedFile()));

			// Obtains first line of parameter file
			String parameters = fileIn.readLine();

			// Parses parameter line
			while (parameters != null) {
				StringTokenizer st = new StringTokenizer(parameters);
				if (st.countTokens() != 2) {
					throw new IOException("Invalid parameter line");
				} else {
					String firstToken = st.nextToken();
					String secondToken = st.nextToken();
					if (firstToken.equals("cellName")) {
						cellName = secondToken;
					} else if (firstToken.equals("electricVersion")) {
						eleVersion = secondToken;
					} else if (firstToken.equals("m2Width")) {
						m2Width = Double.valueOf(secondToken);
					} else if (firstToken.equals("m1Width")) {
						m1Width = Double.valueOf(secondToken);
					} else if (firstToken.equals("polyWidth")) {
						polyWidth = Double.valueOf(secondToken);
					} else if (firstToken.equals("weakPolyWidth")) {
						weakPolyWidth = Double.valueOf(secondToken);
					} else if (firstToken.equals("powerAndGndWide")) {
						powerAndGndWide = Double.valueOf(secondToken);
					} else if (firstToken
							.equals("transistorToActiveConnection")) {
						transistorToActiveConnection = Double
								.valueOf(secondToken);
					} else if (firstToken.equals("m2ToM2Dist")) {
						m2ToM2 = Double.valueOf(secondToken);
					} else if (firstToken.equals("weakTransistorSize")) {
						weakPmosSize = Double.valueOf(secondToken);
					} else if (firstToken.equals("m1ToM1Dist")) {
						m1ToM1 = Double.valueOf(secondToken);
					} else if (firstToken.equals("polyToPolyDist")) {
						polyToPoly = Double.valueOf(secondToken);
					} else if (firstToken.equals("polyToDiffusionDist")) {
						polyToDiff = Double.valueOf(secondToken);
					} else if (firstToken.equals("wellSurround")) {
						wellSurround = Double.valueOf(secondToken);
					} else if (firstToken.equals("activeWidth")) {
						activeWidth = Double.valueOf(secondToken);
					} else if (firstToken.equals("weakActiveWidth")) {
						weakActiveWidth = Double.valueOf(secondToken);
					} else if (firstToken
							.equals("weakTransistorToActiveConnection")) {
						weakTransistorToActiveConnection = Double
								.valueOf(secondToken);
					} else if (firstToken.equals("nMosInvWidth")) {
						nMosInvWidth = Double.valueOf(secondToken);
					} else {
						throw new IOException("Invalid Parameter, \""
								+ firstToken + "\", in txt file");
					}
				}

				parameters = fileIn.readLine();// read-in next line of
				// parameter file
			}
		} catch (IOException e) {
			System.out.println(e);
		}
	}

	public String getCellName() {
		return cellName;
	}

	public String getEleVersion() {
		return eleVersion;
	}

	public double getM1ToM1() {
		return m1ToM1;
	}

	public double getM1Width() {
		return m1Width;
	}

	public double getM2ToM2() {
		return m2ToM2;
	}

	public double getM2Width() {
		return m2Width;
	}

	public double getNMosInvWidth() {
		return nMosInvWidth;
	}

	public double getPolyToDiff() {
		return polyToDiff;
	}

	public double getPolyToPoly() {
		return polyToPoly;
	}

	public double getPolyWidth() {
		return polyWidth;
	}

	public double getPowerAndGndWide() {
		return powerAndGndWide;
	}

	public double getTransistorToActiveConnection() {
		return transistorToActiveConnection;
	}

	public double getWeakPolyWidth() {
		return weakPolyWidth;
	}

	public double getWeakTransistorToActiveConnection() {
		return weakTransistorToActiveConnection;
	}

	public double getWeakActiveWidth() {
		return weakActiveWidth;
	}

	public double getActiveWidth() {
		return activeWidth;
	}

	public double getWellSurround() {
		return wellSurround;
	}

	public double getWeakPmosSize() {
		return weakPmosSize;
	}
}
