package plaGen;

import java.util.Hashtable;

import java.io.*;

import java.util.*;

public class SchematicManager

{

	// Don't use this, use nx() and ny() instead

	public static final int NI_dx = 0;

	public static final int NI_dy = 1;

	public static final int NI_sx = 2;

	public static final int NI_sy = 3;

	public static final int NI_gx = 4;

	public static final int NI_gy = 5;

	public static final int NI_x = 6;

	public static final int NI_y = 7;

	// Inverter Pair Interface

	public static final int INVPAIR_SIZE = 4;

	public static final int INVPAIR_in = 0;

	public static final int INVPAIR_out = 1;

	public static final int INVPAIR_vdd = 2;

	public static final int INVPAIR_gnd = 3;

	// AND Plane Interface

	public static final int ANDPLANE_input = 0;

	public static final int ANDPLANE_output = 1;

	public static final int ANDPLANE_pullup = 2;

	public static final int ANDPLANE_ground = 3;

	// OR Plane Interface

	public static final int ORPLANE_input = 0;

	public static final int ORPLANE_output = 1;

	public static final int ORPLANE_pullup = 2;

	public static final int ORPLANE_ground = 3;

	// AND Plane spacing

	public static final int xloutoff = 10; // how far left to extend the output
											// wire in AND plane

	public static final int xlgndoff = 5; // how far left to extend the ground
											// wire in AND plane

	public static final int xroutoff = 10;

	Hashtable NodeTable;

	Hashtable ArcTable;

	Hashtable ExportTable;

	protected float invXSpacing;

	protected float Xwidth;

	protected float Xheight;

	// transistor sizes

	protected float XS_narray_w = 4;

	protected float XS_narray_h = 2;

	protected float XS_pullup_w = 3;

	protected float XS_pullup_h = 3;

	protected float XS_invn_w = 5;

	protected float XS_invn_h = 2;

	protected float XS_invp_w = 10;

	protected float XS_invp_h = 2;

	protected String cellName;

	protected String libName;

	SchNode myNode;

	SchematicManager(String cname)

	{

		NodeTable = new Hashtable();

		ArcTable = new Hashtable();

		ExportTable = new Hashtable();

		Xwidth = 4;

		Xheight = 4;

		cellName = cname;

		libName = cname;

	}

	public String CreatePMos(float xpos, float ypos)

	{

		return CreatePMos(xpos, ypos, 0);

	}

	public String CreatePMos(float xpos, float ypos, float rot)

	{

		int id = NodeTable.size();

		String name = SchNode.S_pmos + "@" + id;

		SchNode pmos = new SchNode(SchNode.NT_pmos, xpos, ypos, XS_pullup_w,
				XS_pullup_h, rot, name, this);

		NodeTable.put(name, pmos);

		return name;

	}

	public String CreatePMosI(float xpos, float ypos, float rot)

	{

		int id = NodeTable.size();

		String name = SchNode.S_pmos + "@" + id;

		SchNode pmos = new SchNode(SchNode.NT_pmosi, xpos, ypos, XS_invp_w,
				XS_invp_h, rot, name, this);

		NodeTable.put(name, pmos);

		return name;

	}

	public String CreateNMos(float xpos, float ypos)

	{

		return CreateNMos(xpos, ypos, 0);

	}

	public String CreateNMos(float xpos, float ypos, float rot)

	{

		int id = NodeTable.size();

		String name = SchNode.S_nmos + "@" + id;

		SchNode nmos = new SchNode(SchNode.NT_nmos, xpos, ypos, XS_narray_w,
				XS_narray_h, rot, name, this);

		NodeTable.put(name, nmos);

		return name;

	}

	public String CreateNMosI(float xpos, float ypos, float rot)

	{

		int id = NodeTable.size();

		String name = SchNode.S_nmos + "@" + id;

		SchNode nmos = new SchNode(SchNode.NT_nmos, xpos, ypos, XS_invn_w,
				XS_invn_h, rot, name, this);

		NodeTable.put(name, nmos);

		return name;

	}

	public String CreateWire(float hx, float hy, float tx, float ty, String hn,
			String hp, String tn, String tp)

	{

		int id = ArcTable.size();

		String name = SchArc.S_wire + "@" + id;

		SchArc wire = new SchArc(SchArc.AT_wire, 0, hn, hp, hx, hy, tn, tp, tx,
				ty, name, this);

		ArcTable.put(name, wire);

		return name;

	}

	public String CreateWire(String h, String hp, String t, String tp)

	{

		SchNode hn = (SchNode) NodeTable.get(h);

		SchNode tn = (SchNode) NodeTable.get(t);

		float hx = hn.px(hp);

		float hy = hn.py(hp);

		float tx = tn.px(tp);

		float ty = tn.py(tp);

		return CreateWire(hx, hy, tx, ty, h, hp, t, tp);

	}

	public String CreatePin(float xpos, float ypos)

	{

		int id = NodeTable.size();

		String name = SchNode.S_pin + "@" + id;

		SchNode pin = new SchNode(SchNode.NT_pin, xpos, ypos, .5f, .5f, 0,
				name, this);

		NodeTable.put(name, pin);

		return name;

	}

	public String CreateVdd(float xpos, float ypos)

	{

		int id = NodeTable.size();

		String name = SchNode.S_power + "@" + id;

		SchNode power = new SchNode(SchNode.NT_power, xpos, ypos, 3, 3, 0,
				name, this);

		NodeTable.put(name, power);

		return name;

	}

	public String CreateGnd(float xpos, float ypos)

	{

		int id = NodeTable.size();

		String name = SchNode.S_ground + "@" + id;

		SchNode ground = new SchNode(SchNode.NT_ground, xpos, ypos, 3, 4, 0,
				name, this);

		NodeTable.put(name, ground);

		return name;

	}

	public String CreateInputEx(String p, String name)

	{

		int id = ExportTable.size();

		SchExport export = new SchExport(name, p, "I", this);

		ExportTable.put(name, export);

		return null;

	}

	public String CreateOutputEx(String p, String name)

	{

		int id = ExportTable.size();

		SchExport export = new SchExport(name, p, "O", this);

		ExportTable.put(name, export);

		return null;

	}

	public void CreateInverter(float xpos, float ypos)

	{

		// Create Nodes

		String gpin = CreatePin(xpos - 1, ypos + 3);

		String cpin = CreatePin(xpos + 2, ypos + 3);

		String ipin = CreatePin(xpos - 4, ypos + 3);

		String opin = CreatePin(xpos + 5, ypos + 3);

		String nmos = CreateNMos(xpos, ypos, 900);

		String pmos = CreatePMos(xpos, ypos + 6, 900);

		String vdd = CreateVdd(xpos + 2, ypos + 11);

		String gnd = CreateGnd(xpos + 2, ypos - 5);

		// Create Arcs

		String cwirel = CreateWire(nmos, "d", cpin, "");

		String cwireh = CreateWire(cpin, "", pmos, "s");

		String gwireh = CreateWire(nmos, "g", gpin, "");

		String gwirel = CreateWire(gpin, "", pmos, "g");

		String iwire = CreateWire(ipin, "", gpin, "");

		String owire = CreateWire(cpin, "", opin, "");

		String gndwire = CreateWire(gnd, "gnd", nmos, "s");

		String vddwire = CreateWire(pmos, "d", vdd, "");

		// Create Exports

		String input = CreateInputEx(ipin, "a");

		String output = CreateOutputEx(opin, "y");

	}

	public String[] CreateInverterPair(float xpos, float ypos)

	{

		String gpin = CreatePin(xpos - 1, ypos + 3);

		String cpin = CreatePin(xpos + 2, ypos + 3);

		String ipin = CreatePin(xpos - 2, ypos + 3);

		// String opin = CreatePin(xpos+3, ypos+3);

		String nmos = CreateNMosI(xpos, ypos, 900);

		String pmos = CreatePMosI(xpos, ypos + 6, 900);

		String vddpin = CreatePin(xpos + 2, ypos + 9);

		String gndpin = CreatePin(xpos + 2, ypos - 3);

		// Create Arcs

		// source/drain arcs

		String cwirel = CreateWire(nmos, "d", cpin, "");

		String cwireh = CreateWire(cpin, "", pmos, "s");

		String gwireh = CreateWire(nmos, "g", gpin, "");

		String gwirel = CreateWire(gpin, "", pmos, "g");

		String iwire = CreateWire(ipin, "", gpin, "");

		// String owire = CreateWire(cpin, "", opin, "");

		String vddwire = CreateWire(pmos, "d", vddpin, "");

		String gndwire = CreateWire(gndpin, "", nmos, "s");

		String[] out = new String[INVPAIR_SIZE];

		out[INVPAIR_in] = ipin;

		out[INVPAIR_out] = cpin;

		out[INVPAIR_vdd] = vddpin;

		out[INVPAIR_gnd] = gndpin;

		return out;

	}

	public String[] CreateInverterPairOutput(float xpos, float ypos)

	{

		String gpin = CreatePin(xpos - 1, ypos + 3);

		String cpin = CreatePin(xpos + 2, ypos + 3);

		String ipin = CreatePin(xpos - 2, ypos + 3);

		String opin = CreatePin(xpos + 3, ypos + 3);

		String nmos = CreateNMosI(xpos, ypos, 900);

		String pmos = CreatePMosI(xpos, ypos + 6, 900);

		String vddpin = CreatePin(xpos + 2, ypos + 9);

		String gndpin = CreatePin(xpos + 2, ypos - 3);

		// Create Arcs

		// source/drain arcs

		String cwirel = CreateWire(nmos, "d", cpin, "");

		String cwireh = CreateWire(cpin, "", pmos, "s");

		String gwireh = CreateWire(nmos, "g", gpin, "");

		String gwirel = CreateWire(gpin, "", pmos, "g");

		String iwire = CreateWire(ipin, "", gpin, "");

		String owire = CreateWire(cpin, "", opin, "");

		String vddwire = CreateWire(pmos, "d", vddpin, "");

		String gndwire = CreateWire(gndpin, "", nmos, "s");

		String[] out = new String[INVPAIR_SIZE];

		out[INVPAIR_in] = ipin;

		out[INVPAIR_out] = opin;

		out[INVPAIR_vdd] = vddpin;

		out[INVPAIR_gnd] = gndpin;

		return out;

	}

	public Vector CreateANDPlane(float xpos, float ypos, int[][] plane,
			int numInputs, int numCases)

	{

		int xoffset = 0;

		int yoffset = 0;

		int xspacing = 10;

		int yspacingeven = 4;

		int yspacingodd = 8;

		int notOffset = 5;

		int topgnd = (int) Math.ceil((double) numCases / 2.0);

		int numgndpins = numCases / 2 + numCases % 2;

		Vector ANDPlaneInterface = new Vector();

		String[] groundpins = new String[numgndpins];

		String[] inputpins = new String[numInputs * 2];

		String[] pulluppins = new String[numCases];

		String[] outputpins = new String[numCases];

		// Create nMos Transistors

		String[][] nmos = new String[numCases][numInputs];

		int evod = 1;

		for (int r = numCases - 1; r >= 0; r--)

		{

			xoffset = 0;

			for (int c = 0; c < numInputs; c++)

			{

				if (plane[r][c] == 0)

				{

					nmos[r][c] = CreateNMos(xpos + xoffset, ypos + yoffset, 900);

				}

				else if (plane[r][c] == 1)

				{

					nmos[r][c] = CreateNMos(xpos + xoffset + notOffset, ypos
							+ yoffset, 900);

				}

				else if (plane[r][c] == 2)

				{

					nmos[r][c] = null;

				}

				else

				{

					System.out.println("ERROR: Bad AND Plane. Unknown symbol "
							+ plane[r][c]);

					System.exit(1);

				}

				xoffset = xoffset + xspacing;

			}

			if (evod % 2 == 0) {

				yoffset = yoffset + yspacingodd;

			}

			else {

				yoffset = yoffset + yspacingeven;

			}

			evod++;

		}

		// Wire gates

		String prv = null;

		for (int c = 0; c < numInputs; c++)

		{

			// search for nots

			for (int r = 0; r < numCases; r++)

			{

				if (plane[r][c] == 1)

				{

					if (prv != null)

					{

						CreateWire(prv, "g", nmos[r][c], "g");

					}

					prv = nmos[r][c];

				}

			}

			if (prv == null)

			{

				inputpins[2 * c + 1] = null;

			}

			else

			{

				inputpins[2 * c + 1] = CreatePin(nx(prv, "g"), ny(prv, "g"));

				CreateWire(inputpins[2 * c + 1], "", prv, "g");

			}

			prv = null;

			// search for non-nots

			for (int r = 0; r < numCases; r++)

			{

				if (plane[r][c] == 0)

				{

					if (prv != null)

					{

						CreateWire(prv, "g", nmos[r][c], "g");

					}

					prv = nmos[r][c];

				}

			}

			if (prv == null)

			{

				inputpins[2 * c] = null;

			}

			else

			{

				inputpins[2 * c] = CreatePin(nx(prv, "g"), ny(prv, "g"));

				CreateWire(inputpins[2 * c], "", prv, "g");

			}

			prv = null;

		}

		// Wire source and drain

		prv = null;

		String prvdn = null;

		evod = 1;

		for (int r = numCases - 1; r >= 0; r--)

		{

			String pin = null;

			String pingnd = null;

			String pinout = null;

			String out = null;

			for (int c = 0; c < numInputs; c++)

			{

				String pindn = null;

				if (evod % 2 == 0) // creating drain (top) and ground wires

				{

					if ((nmos[r][c] != null) && (nmos[r + 1][c] != null)) // two
																			// nmos's
																			// share
																			// ground
																			// in
																			// the
																			// column

					{

						if (plane[r][c] < plane[r + 1][c]) // top t leads

						{

							pindn = CreatePin(nx(nmos[r + 1][c], "s"), ny(
									nmos[r + 1][c], "d"));

							if (prvdn != null)

							{

								CreateWire(prvdn, "", nmos[r][c], "s");

							}

							else // create out pin for GND

							{

								pingnd = CreatePin(xpos - xlgndoff, ny(
										nmos[r][c], "s"));
								;

								CreateWire(pingnd, "", nmos[r][c], "s");

							}

						}

						else // bot t leads

						{

							pindn = CreatePin(nx(nmos[r][c], "s"), ny(
									nmos[r][c], "s"));

							if (prvdn != null)

							{

								CreateWire(prvdn, "", nmos[r + 1][c], "d");

							}

							else // create out pin for GND

							{

								pingnd = CreatePin(xpos - xlgndoff, ny(
										nmos[r + 1][c], "d"));

								CreateWire(pingnd, "", nmos[r + 1][c], "d");

							}

						}

						if (prv != null)

						{

							String prvt = CreatePin(nx(nmos[r][c], "d"), ny(
									nmos[r][c], "d"));

							CreateWire(prv, "", prvt, "");

							CreateWire(nmos[r][c], "d", prvt, "");

							prv = prvt;

							// CreateWire(prv, "d", nmos[r][c], "d");

						}

						else // wire drain (top) to out pin

						{

							pinout = CreatePin(xpos - xloutoff, ny(nmos[r][c],
									"d"));

							String prvt = CreatePin(nx(nmos[r][c], "d"), ny(
									nmos[r][c], "d"));

							CreateWire(pinout, "", prvt, "");

							CreateWire(nmos[r][c], "d", prvt, "");

							prv = prvt;

							// CreateWire(pinout, "", nmos[r][c], "d");

						}

						// remove this

						// prv = nmos[r][c];

						CreateWire(nmos[r][c], "s", pindn, "");

						CreateWire(nmos[r + 1][c], "d", pindn, "");

						prvdn = pindn;

					}

					else // only one nmos on ground in this column

					{

						if (nmos[r][c] != null) // upper nmos present

						{

							pindn = CreatePin(nx(nmos[r][c], "s"), ny(
									nmos[r][c], "s"));

							CreateWire(pindn, "", nmos[r][c], "s");

							if (prv != null) // wire drain (top) to last nmos

							{

								String prvt = CreatePin(nx(nmos[r][c], "d"),
										ny(nmos[r][c], "d"));

								CreateWire(prv, "", prvt, "");

								CreateWire(nmos[r][c], "d", prvt, "");

								prv = prvt;

								// CreateWire(prv, "d", nmos[r][c], "d");

							}

							else // this is first drain to wire so wire to
									// extension

							{

								pinout = CreatePin(xpos - xloutoff, ny(
										nmos[r][c], "d"));

								prv = CreatePin(nx(nmos[r][c], "d"), ny(
										nmos[r][c], "d"));

								CreateWire(pinout, "", prv, "");

								CreateWire(nmos[r][c], "d", prv, "");

								// CreateWire(pinout, "", nmos[r][c], "d");

							}

							// remove this

							// prv = nmos[r][c];

						}

						if (nmos[r + 1][c] != null) // lower nmos present

						{

							if (pindn == null)

							{

								pindn = CreatePin(nx(nmos[r + 1][c], "d"), ny(
										nmos[r + 1][c], "d"));

							}

							CreateWire(nmos[r + 1][c], "d", pindn, "");

						}

						if (pindn != null)

						{

							if (prvdn != null)

							{

								CreateWire(prvdn, "", pindn, "");

							}

							else // create out pin for GND

							{

								pingnd = CreatePin(xpos - xlgndoff, ny(pindn,
										""));

								CreateWire(pingnd, "", pindn, "");

							}

							prvdn = pindn;

						}

					}

				}

				else // creating only source output wire

				{

					if (nmos[r][c] != null)

					{

						if (prv != null)

						{

							String prvt = CreatePin(nx(nmos[r][c], "s"), ny(
									nmos[r][c], "s"));

							CreateWire(nmos[r][c], "s", prvt, "");

							CreateWire(prv, "", prvt, "");

							prv = prvt;

							// CreateWire(prv, "s", nmos[r][c], "s");

						}

						else

						{

							pinout = CreatePin(xpos - xloutoff, ny(nmos[r][c],
									"s"));

							prv = CreatePin(nx(nmos[r][c], "s"), ny(nmos[r][c],
									"s"));

							CreateWire(pinout, "", prv, "");

							CreateWire(nmos[r][c], "s", prv, "");

							// CreateWire(pinout, "", nmos[r][c], "s");

						}

						// remove this

						// prv = nmos[r][c];

					}

				}

			}

			if (prv == null) // FIX?

			{

				System.out
						.println("ERROR: unnecessary case in AND plane on row:"
								+ r);

				System.exit(1);

			}

			if (evod % 2 == 0)

			{

				groundpins[(evod / 2) - 1] = pingnd;

				// wire last nmos drain to output

				// outputpins[r] = CreatePin(nx(prv, "d"), ny(prv, "d"));

				// CreateWire(prv, "d", outputpins[r], "");

				// this is correct line

				outputpins[r] = prv;

			}

			else

			{

				// wire last nmos source to output

				// outputpins[r] = CreatePin(nx(prv, "s"), ny(prv, "s"));

				// CreateWire(prv, "s", outputpins[r], "");

				// this is correct line

				outputpins[r] = prv;

			}

			pulluppins[r] = pinout;

			if ((r == 0) && (evod % 2 != 0)) // need to wire top ground plane
												// still

			{

				String gpin = null;

				prv = null;

				for (int c = 0; c < numInputs; c++)

				{

					if (nmos[r][c] != null)

					{

						if (gpin == null)

						{

							gpin = CreatePin(xpos - xlgndoff, ny(nmos[r][c],
									"d"));

							CreateWire(gpin, "", nmos[r][c], "d");

							groundpins[(evod / 2)] = gpin;

						}

						if (prv != null)

						{

							CreateWire(prv, "d", nmos[r][c], "d");

						}

						prv = nmos[r][c];

					}

				}

			}

			prv = null;

			prvdn = null;

			evod++;

		}

		// Create Ground Wiring

		for (int i = groundpins.length - 1; i > 0; i--)

		{

			CreateWire(groundpins[i], "", groundpins[i - 1], "");

		}

		String sgnd = CreateGnd(xpos - xlgndoff, ypos - 8);

		CreateWire(sgnd, "gnd", groundpins[0], "");

		// Create Interface

		ANDPlaneInterface.add(ANDPLANE_input, inputpins);

		ANDPlaneInterface.add(ANDPLANE_output, outputpins);

		ANDPlaneInterface.add(ANDPLANE_pullup, pulluppins);

		ANDPlaneInterface.add(ANDPLANE_ground, groundpins);

		return ANDPlaneInterface;

	}

	public void WireANDtoInv(Vector APi, String[] INVi)

	{

		String[] ANDi = (String[]) APi.elementAt(ANDPLANE_input);

		if (ANDi.length != INVi.length)

		{

			System.out
					.println("ERROR: number of pins between AND Plane and Input Inverters do no match");

			System.exit(1);

		}

		for (int i = 0; i < ANDi.length; i++)

		{

			if (ANDi[i] != null)

			{

				if (i % 2 != 0)

				{

					String outpin = CreatePin(nx(INVi[i], "") + 1, ny(INVi[i],
							""));

					CreateWire(INVi[i], "", outpin, "");

					CreateWire(ANDi[i], "", outpin, "");

				}

				else

				{

					CreateWire(ANDi[i], "", INVi[i], "");

				}

			}

		}

	}

	public Vector CreateORPlane(float xpos, float ypos, int[][] plane,
			int numOutputs, int numCases)

	{

		String[][] nmos = new String[numCases][numOutputs];

		Vector ORPlaneInterface = new Vector();
		;

		int yoffset = 0;

		int xoffset = 0;

		int xspacingodd = 10;

		int xspacingeven = 4;

		int yspacingeven = 8;

		int yspacingodd = 4;

		int xspacing = 8;

		int inpinoff = 5;

		int outpinoff = 10;

		int gndpinoff = 6;

		int xgnd = 4;

		int ygnd = 8;

		String[] inpins = new String[numCases];

		String[] outpins = new String[numOutputs];

		String[] pulluppins = new String[numOutputs];

		String[] groundpins = new String[numOutputs];

		// Create Transistors

		int evod = 1;

		for (int r = numCases - 1; r >= 0; r--)

		{

			for (int c = 0; c < numOutputs; c++)

			{

				if (plane[r][c] == 1)

				{

					nmos[r][c] = CreateNMos(xpos + xoffset, ypos + yoffset, 0);

				}

				else

				{

					nmos[r][c] = null;

				}

				/*
				 * if(c%2 == 0) { xoffset = xoffset + xspacingeven; } else {
				 * xoffset = xoffset + xspacingodd; }
				 */

				xoffset += xspacing;

			}

			evod++;

			if (evod % 2 == 0)

			{

				yoffset = yoffset + yspacingeven;

			}

			else

			{

				yoffset = yoffset + yspacingodd;

			}

			xoffset = 0;

		}

		// wire gates

		String prv = null;

		for (int r = numCases - 1; r >= 0; r--)

		{

			prv = null;

			for (int c = 0; c < numOutputs; c++)

			{

				if (nmos[r][c] != null)

				{

					if (prv != null)

					{

						CreateWire(prv, "g", nmos[r][c], "g");

					}

					else

					{

						inpins[r] = CreatePin(xpos - inpinoff, ny(nmos[r][c],
								"g"));

						CreateWire(inpins[r], "", nmos[r][c], "g");

					}

					prv = nmos[r][c];

				}

			}

		}

		// wire outpus

		prv = null;

		for (int c = 0; c < numOutputs; c++)

		{

			String p = "d";

			/*
			 * if(c%2 == 0) { p = "s"; }
			 */

			prv = null;

			for (int r = numCases - 1; r >= 0; r--)

			{

				if (nmos[r][c] != null)

				{

					if (prv != null)

					{

						CreateWire(prv, p, nmos[r][c], p);

					}

					else

					{

						outpins[c] = CreatePin(nx(nmos[r][c], p), ny(
								nmos[r][c], p)); // ypos - outpinoff

						CreateWire(outpins[c], "", nmos[r][c], p);

					}

					prv = nmos[r][c];

				}

			}

			if (prv == null) // FIXE ME

			{

				int port = (numOutputs - 1) - c;

				System.out.println("WARNING: unnecessary output on port : "
						+ port);

				// System.exit(1);

				// Create output pin and pullup pin

				String hpin = CreatePin(2 + xpos + xoffset + c * xspacing, ypos
						+ yoffset);

				outpins[c] = hpin;

				pulluppins[c] = hpin;

			}

			else

			{

				pulluppins[c] = CreatePin(nx(prv, p), ny(prv, p));

				CreateWire(pulluppins[c], "", prv, p);

			}

		}

		// wire grounds

		for (int c = 0; c < numOutputs; c++)

		{

			String p = "s";

			prv = null;

			for (int r = 0; r < numCases; r++)

			// for(int r = numCases-1; r >= 0; r--)

			{

				/*
				 * if( prv == null ) { groundpins[c] = CreatePin(-2 + xpos +
				 * xoffset + c*xspacing, ypos - gndpinoff); } if(nmos[r][c] !=
				 * null) { if( prv != null ) { CreateWire(prv, p, nmos[r][c],
				 * p); } else { //groundpins[c] = CreatePin(nx(nmos[r][c], p),
				 * ypos - gndpinoff); CreateWire(groundpins[c], "", nmos[r][c],
				 * p); } prv = nmos[r][c]; }
				 */

				if (nmos[r][c] != null)

				{

					if (prv != null)

					{

						CreateWire(prv, p, nmos[r][c], p);

					}

					prv = nmos[r][c];

				}

			}

			if (prv != null)

			{

				groundpins[c] = CreatePin(-2 + xpos + xoffset + c * xspacing,
						ypos - gndpinoff);

				CreateWire(groundpins[c], "", prv, p);

			}

		}

		// Create Ground Wiring

		/*
		 * for( int i = groundpins.length-1; i > 0; i--) {
		 * CreateWire(groundpins[i], "", groundpins[i-1], ""); } String sgnd =
		 * CreateGnd(xpos-xgnd, ypos-ygnd); CreateWire(sgnd, "gnd",
		 * groundpins[0], "");
		 */

		String sgnd = CreateGnd(xpos - xgnd, ypos - ygnd);

		prv = sgnd;

		boolean atgnd = true;

		for (int i = 0; i < groundpins.length; i++)

		{

			if (groundpins[i] != null)

			{

				if (atgnd)

				{

					CreateWire(prv, "gnd", groundpins[i], "");

					prv = groundpins[i];

					atgnd = false;

				}

				else

				{

					CreateWire(prv, "", groundpins[i], "");

					prv = groundpins[i];

				}

			}

			// CreateWire(groundpins[i], "", groundpins[i-1], "");

		}

		ORPlaneInterface.add(ORPLANE_input, inpins);

		ORPlaneInterface.add(ORPLANE_output, outpins);

		ORPlaneInterface.add(ORPLANE_pullup, pulluppins);

		ORPlaneInterface.add(ORPLANE_ground, groundpins);

		return ORPlaneInterface;

	}

	public void WireORtoAND(Vector ORi, Vector ANDi)

	{

		String[] ANDout = (String[]) ANDi.elementAt(ANDPLANE_output);

		String[] ORin = (String[]) ORi.elementAt(ORPLANE_input);

		if (ORin.length != ANDout.length)

		{

			System.out
					.println("ERROR: number of pins between AND Plane and OR Plane do no match");

			System.exit(1);

		}

		for (int i = 0; i < ORin.length; i++)

		{

			if (ORin[i] == null)

			{

				int cs = i + 1;

				System.out.println("Warning: OR plane does not use case " + cs);

				// System.exit(1);

			}

			if (ORin[i] != null) {

				CreateWire(ORin[i], "", ANDout[i], "");

			}

		}

	}

	public String[] CreateOutputInverters(float xpos, float ypos,
			String outputBus, int busWidth)

	{

		int offset = 0;

		int invSpacing = 8;

		// Create Transistors

		Vector invPairs = new Vector();

		for (int i = 0; i < busWidth; i++)

		{

			invPairs.add(CreateInverterPairOutput(xpos + offset, ypos));

			offset = (i + 1) * invSpacing;

		}

		// Connect VDD and GND

		for (int i = 0; i < busWidth - 1; i++)

		{

			String[] inv = (String[]) invPairs.elementAt(i);

			String[] invNext = (String[]) invPairs.elementAt(i + 1);

			CreateWire(inv[INVPAIR_vdd], "", invNext[INVPAIR_vdd], "");

			CreateWire(inv[INVPAIR_gnd], "", invNext[INVPAIR_gnd], "");

			if (i == busWidth - 2)

			{

				String vdd = CreateVdd(nx(invNext[INVPAIR_vdd], ""), ny(
						invNext[INVPAIR_vdd], "") + 2);

				String gnd = CreateGnd(nx(invNext[INVPAIR_gnd], ""), ny(
						invNext[INVPAIR_gnd], "") - 2);

				CreateWire(gnd, "gnd", invNext[INVPAIR_gnd], "");

				CreateWire(invNext[INVPAIR_vdd], "", vdd, "");

			}

		}

		// Create Output Exports

		for (int i = 0; i < busWidth; i++)

		{

			String[] inv = (String[]) invPairs.elementAt(i);

			int busi = busWidth - (i + 1);

			CreateOutputEx(inv[INVPAIR_out], outputBus + "[" + busi + "]");

		}

		String[] out = new String[busWidth];

		for (int i = 0; i < busWidth; i++)

		{

			String[] inv = (String[]) invPairs.elementAt(i);

			out[i] = inv[INVPAIR_in];

			// out[2*i+1] = inv[INVPAIR_out];

		}

		return out;

	}

	public void WireORtoInv(Vector ORi, String[] INVi)

	{

		String[] ORout = (String[]) ORi.elementAt(ORPLANE_output);

		if (ORout.length != INVi.length)

		{

			System.out
					.println("ERROR: number of pins between OR Plane and Input Inverters do no match");

			System.exit(1);

		}

		for (int i = 0; i < ORout.length; i++)

		{

			CreateWire(INVi[i], "", ORout[i], "");

		}

	}

	public String[] CreateOrPullUp(float xpos, float ypos, int numOut)

	{

		int spacing = 8;

		int offset = 0;

		int vddxoff = 2;

		int vddyoff = 4;

		int gndxoff = 6;

		int gndyoff = 2;

		String[] inpins = new String[numOut];

		String[] pmos = new String[numOut];

		// Create PMos

		for (int i = 0; i < numOut; i++)

		{

			pmos[i] = CreatePMos(xpos + offset, ypos, 900);

			inpins[i] = CreatePin(nx(pmos[i], "s"), ny(pmos[i], "s"));

			CreateWire(inpins[i], "", pmos[i], "s");

			offset = (i + 1) * spacing;

		}

		// Create Power

		for (int i = 0; i < pmos.length - 1; i++)

		{

			if (i == 0)

			{

				String vdd = CreateVdd(xpos + vddxoff, ypos + vddyoff);

				CreateWire(vdd, "", pmos[i], "d");

			}

			CreateWire(pmos[i], "d", pmos[i + 1], "d");

		}

		// Create Gate Input

		for (int i = 0; i < pmos.length - 1; i++)

		{

			if (i == 0)

			{

				String gnd = CreateGnd(xpos - gndxoff, ypos - gndyoff);

				CreateWire(gnd, "gnd", pmos[i], "g");

			}

			CreateWire(pmos[i], "g", pmos[i + 1], "g");

		}

		return inpins;

	}

	public void WireORtoPullUp(Vector ORi, String[] PUi)

	{

		String[] ORpullup = (String[]) ORi.elementAt(ORPLANE_pullup);

		if (ORpullup.length != PUi.length)

		{

			System.out
					.println("ERROR: number of pins between OR Plane and Pull-Ups do no match");

			System.exit(1);

		}

		for (int i = 0; i < ORpullup.length; i++)

		{

			CreateWire(ORpullup[i], "", PUi[i], "");

		}

	}

	public String[] CreateAndPullUp(float xpos, float ypos, int numCases)

	{

		int spacing = 8;

		int spacingeven = 4;

		int spacingodd = 8;

		int offset = (int) ypos;

		int vddxoff = -2;

		int vddyoff = 4;

		int gndxoff = 0;

		int gndyoff = -6;

		String[] inpins = new String[numCases];

		String[] pmos = new String[numCases];

		// Create PMos

		int evod = numCases;

		for (int i = 0; i < numCases; i++)

		{

			pmos[i] = CreatePMos(xpos, offset, 0);

			inpins[i] = CreatePin(nx(pmos[i], "d"), ny(pmos[i], "d"));

			CreateWire(inpins[i], "", pmos[i], "d");

			evod++;

			if (evod % 2 == 0) {

				offset = offset - spacingeven;

			}

			else {

				offset = offset - spacingodd;

			}

		}

		// Create Power

		for (int i = 0; i < pmos.length - 1; i++)

		{

			if (i == 0)

			{

				String vdd = CreateVdd(xpos + vddxoff, ypos + vddyoff);

				CreateWire(vdd, "", pmos[i], "s");

			}

			CreateWire(pmos[i], "s", pmos[i + 1], "s");

		}

		// Create Gate Input

		for (int i = 0; i < pmos.length - 1; i++)

		{

			CreateWire(pmos[i], "g", pmos[i + 1], "g");

			if (i == pmos.length - 2)

			{

				String gnd = CreateGnd(xpos + gndxoff, ny(pmos[i + 1], "")
						+ gndyoff);

				CreateWire(gnd, "gnd", pmos[i + 1], "g");

			}

		}

		return inpins;

	}

	public void WireANDtoPullUp(Vector ANDi, String[] PUi)

	{

		String[] ANDpullup = (String[]) ANDi.elementAt(ANDPLANE_pullup);

		if (ANDpullup.length != PUi.length)

		{

			System.out
					.println("ERROR: number of pins between AND Plane and Pull-Ups do no match");

			System.exit(1);

		}

		for (int i = 0; i < ANDpullup.length; i++)

		{

			CreateWire(ANDpullup[i], "", PUi[i], "");

		}

	}

	public String[] CreateInputInverters(float xpos, float ypos,
			String inputBus, int busWidth)

	{

		int offset = 0;

		int invSpacing = 10;

		// Create Transistors

		Vector invPairs = new Vector();

		for (int i = 0; i < busWidth; i++)

		{

			invPairs.add(CreateInverterPair(xpos + offset, ypos));

			offset = (i + 1) * invSpacing;

		}

		// Connect VDD and GND

		for (int i = 0; i < busWidth - 1; i++)

		{

			String[] inv = (String[]) invPairs.elementAt(i);

			String[] invNext = (String[]) invPairs.elementAt(i + 1);

			CreateWire(inv[INVPAIR_vdd], "", invNext[INVPAIR_vdd], "");

			CreateWire(inv[INVPAIR_gnd], "", invNext[INVPAIR_gnd], "");

			if (i == 0)

			{

				String vdd = CreateVdd(xpos + 2, ypos + 11);

				String gnd = CreateGnd(xpos + 2, ypos - 5);

				CreateWire(gnd, "gnd", inv[INVPAIR_gnd], "");

				CreateWire(inv[INVPAIR_vdd], "", vdd, "");

			}

		}

		// Create Input Exports

		for (int i = 0; i < busWidth; i++)

		{

			String[] inv = (String[]) invPairs.elementAt(i);

			int busi = busWidth - (i + 1);

			CreateInputEx(inv[INVPAIR_in], inputBus + "[" + busi + "]");

		}

		String[] out = new String[busWidth * 2];

		for (int i = 0; i < busWidth; i++)

		{

			String[] inv = (String[]) invPairs.elementAt(i);

			out[2 * i] = inv[INVPAIR_in];

			out[2 * i + 1] = inv[INVPAIR_out];

		}

		return out;

	}

	public float nx(String n, String p)

	{

		SchNode N = (SchNode) NodeTable.get(n);

		return N.px(p);

	}

	public float ny(String n, String p)

	{

		SchNode N = (SchNode) NodeTable.get(n);

		return N.py(p);

	}

	private float GetNInfo(String tname, int Info)

	{

		SchNode x = (SchNode) NodeTable.get(tname);

		switch (Info)

		{

		case NI_dx:
			return x.dx();

		case NI_dy:
			return x.dy();

		case NI_sx:
			return x.sx();

		case NI_sy:
			return x.sy();

		case NI_gx:
			return x.gx();

		case NI_gy:
			return x.gy();

		case NI_x:
			return x.x();

		case NI_y:
			return x.y();

		default:
			return -1;

		}

	}

	// Write Functions

	public void WriteLibrary(String filename)

	{

		try

		{

			PrintStream out = new PrintStream((new FileOutputStream(filename)));

			// Write Headers

			WriteLibraryHeader(out);

			WriteCellHeader(out);

			// Write Cell Body

			SchNode node;

			for (Enumeration e = NodeTable.elements(); e.hasMoreElements();)

			{

				node = (SchNode) e.nextElement();

				out.print(node.GetString() + "\n");

			}

			SchArc arc;

			for (Enumeration e = ArcTable.elements(); e.hasMoreElements();)

			{

				arc = (SchArc) e.nextElement();

				out.print(arc.GetString() + "\n");

			}

			SchExport ex;

			for (Enumeration e = ExportTable.elements(); e.hasMoreElements();)

			{

				ex = (SchExport) e.nextElement();

				out.print(ex.GetString() + "\n");

			}

			// Write Cell Footer

			WriteCellFooter(out);

			// Write Library Footer

			WriteLibraryFooter(out);

			// Close file

			out.close();

			System.out.println("Succesfully wrote library file: " + filename);

		}

		catch (IOException e)

		{

			System.out.println(e);

		}

	}

	private void WriteLibraryHeader(PrintStream out)

	{

		out.print("# header information:\n");

		out.print("H" + libName + "|8.04\n\n");

		out.print("# Views:\n");

		out.print("Vschematic|sch\n\n");

		out.print("# Technologies:\n");

		out
				.print("Tmocmos|MoCMOSAlternateActivePolyRules()I1|MoCMOSNumberOfMetalLayers()I3|MoCMOSSecondPolysilicon()I0|ScaleFORmocmos()D300.0\n\n");

	}

	private void WriteCellHeader(PrintStream out)

	{

		out.print("# Cell " + cellName + ";1{sch}\n");

		out.print("C" + cellName
				+ ";1{sch}||schematic|1173062406946|1173062415884|\n");

	}

	private void WriteCellFooter(PrintStream out)

	{

		out.print("X\n\n");

	}

	private void WriteLibraryFooter(PrintStream out)

	{

		out.print("# Groups:\n");

		out.print("G" + cellName + ";1{sch}\n");

	}

	public String text()

	{

		String libText = "";

		// Write Cell Header

		libText = libText + "# Cell " + cellName + ";1{sch}\n";

		libText = libText + "C" + cellName
				+ ";1{sch}||schematic|1173062406946|1173062415884|\n";

		// Write Cell Body

		SchNode node;

		for (Enumeration e = NodeTable.elements(); e.hasMoreElements();)

		{

			node = (SchNode) e.nextElement();

			// out.print(node.GetString()+"\n");

			libText = libText + node.GetString() + "\n";

		}

		SchArc arc;

		for (Enumeration e = ArcTable.elements(); e.hasMoreElements();)

		{

			arc = (SchArc) e.nextElement();

			// out.print(arc.GetString()+"\n");

			libText = libText + arc.GetString() + "\n";

		}

		SchExport ex;

		for (Enumeration e = ExportTable.elements(); e.hasMoreElements();)

		{

			ex = (SchExport) e.nextElement();

			// out.print(ex.GetString()+"\n");

			libText = libText + ex.GetString() + "\n";

		}

		// Write Cell Footer

		libText = libText + "X\n\n";

		return libText;

	}

}