package plaGen;

import java.io.FileOutputStream;
import java.io.PrintStream;

public class plaGen {

	/*
	 * Method: Main Inputs: String[] "args": unused Returns: Void Function: 1.
	 * Creates Paramgetters and Casinfo objects to parse and make available
	 * parameters and case-statement information. 2. Builds the electric .jelib
	 * file by creating a header, generating a cell and calling its "text"
	 * method to get the body of the file, and then joining the header and body.
	 * 3. Calls the outputToFile method to output the jelib file.
	 */
	public static void main(String[] args) {

		Paramgetter paramGetter = new Paramgetter();
		Caseinfo caseInfo = new Caseinfo();
		String headerText = makeHeader(paramGetter);

		String layoutText = new LayCell(caseInfo, paramGetter).text();
		String schematicText = new SchematicCell(caseInfo, paramGetter).text();
		String groupText = makeGroup(paramGetter);
		// System.out.println("1");
		String jelibfile = headerText + schematicText + layoutText + groupText;

		outputToFile(jelibfile, paramGetter);
		// caseInfoTester(caseInfo);
	}

	private static String makeGroup(Paramgetter paramGetter) {
		String name = paramGetter.getCellName();
		return "# Groups:"+"\nG"+name+"_Cell;1{sch}|"+name+"_Cell;1{lay}";
	}

	/*
	 * Method: caseInfoTester Inputs: Caseinfo "caseInfo": Should contain all
	 * pertenant information. Returns: Void Function: Testing method used to
	 * check case-statement parsing Outputs information to the console,
	 * including: 1. name of the input bus 2. number of input bits (n) 3. number
	 * of output bits (m) 4. name of output bus 5. number of case lines,
	 * excluding default (p) 6. the and-matrix (p x n matrix used to generate
	 * the AND plane ) 7. the or-matrix (p x m matrix used to generate the OR
	 * plane)
	 */
	@SuppressWarnings("unused")
	private static void caseInfoTester(Caseinfo caseInfo) {
		int numCases = caseInfo.getNumCases();
		int numInputBits = caseInfo.getNumInputBits();
		int numOutputBits = caseInfo.getNumOutputBits();

		System.out.println("inBusName " + caseInfo.getInBusName());
		System.out.println("numInputBits " + numInputBits);
		System.out.println("numOutputBits " + numOutputBits);
		System.out.println("outBusName " + caseInfo.getOutBusName());
		System.out.println("numCases " + numCases);

		String outLine;

		System.out.println("InputMatrix");
		for (int r = 0; r < numCases; r++) {
			outLine = "";

			for (int c = 0; c < numInputBits; c++) {
				outLine = outLine
						+ String.valueOf(caseInfo.getInputMatrix()[r][c]) + " ";
			}

			System.out.println(outLine);
		}

		System.out.println("OutputMatrix");
		for (int r = 0; r < numCases; r++) {
			outLine = "";

			for (int c = 0; c < numOutputBits; c++) {
				outLine = outLine
						+ String.valueOf(caseInfo.getOutputMatrix()[r][c])
						+ " ";
			}

			System.out.println(outLine);
		}
	}

	/*
	 * Method: outputToFile Inputs: String "jelibfile": Contains the text data
	 * which makes up the jelib file Paramgetter "paramGetter": Used to obtain
	 * the filename. Returns: Void Tasks: Takes in the jelib file information
	 * and the filename, and outputs a .jelib file to the root directory.
	 */
	private static void outputToFile(String jelibfile, Paramgetter paramGetter) {
		try {

			PrintStream outStream = new PrintStream(new FileOutputStream(
					paramGetter.getCellName() + ".jelib"));
			outStream.print(jelibfile);
			outStream.close();
		} catch (Exception e) {
			System.err.println("Error writing to file");
		}

	}

	/*
	 * Method: makeHeader Inputs: Paramgetter "paramgetter": Used to supply
	 * necessary parameters Returns: String consisting of the file's header
	 * information Note:It is anticipated that many of the "magic number"-type
	 * statements (such as, for example "mocmos"...) might be set in a different
	 * way. To this end, such fields have been partitioned out for easy access.
	 */
	private static String makeHeader(Paramgetter paramgetter) {
		String header = null;
		String cellname = paramgetter.getCellName();
		String electricVersion = paramgetter.getEleVersion();
		header = "# header information:\n"
				+ "H"
				+ cellname
				+ "|"
				+ electricVersion
				+ "\n"
				+ "\n"
				+ "# Views:\n"
				// icon generation not yet implemented
				// +"V"+"icon"+"|"+"ic"+"\n"
				+ "V" + "layout" + "|" + "lay" + "\n" + "V" + "schematic" + "|"
				+ "sch\n" + "\n" + "# Technologies:\n" + "T" + "mocmos" + "|"
				+ "MoCMOSAlternateActivePolyRules()I1" + "|"
				+ "MoCMOSNumberOfMetalLayers()I3" + "|"
				+ "MoCMOSSecondPolysilicon()I0" + "|"
				+ "ScaleFORmocmos()D300.0" + "\n" + "\n";

		return header;
	}
}
